# -*- coding: utf-8 -*-
#
# util.py - Misc utilities
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os
import shutil
import subprocess
import urllib.request

import bs4
from cppman import environ

# User-Agent header value to use with all requests
_USER_AGENT = "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/111.0"

def update_mandb_path():
    """Add $XDG_CACHE_HOME/cppman/man to $HOME/.manpath"""
    manpath_file = os.path.join(environ.HOME, ".manpath")
    man_dir = environ.cache_dir
    manindex_dir = environ.manindex_dir

    lines = []

    """ read all lines """
    try:
        with open(manpath_file, 'r') as f:
            lines = f.readlines()
    except IOError:
        return

    """ remove MANDATORY_MANPATH and MANDB_MAP entry """
    lines = [line for line in lines if man_dir not in line]

    with open(manpath_file, 'w') as f:
        if environ.config.UpdateManPath:
            lines.append('MANDATORY_MANPATH\t%s\n' % man_dir)
            lines.append('MANDB_MAP\t\t\t%s\t%s\n' % (man_dir, manindex_dir))

        f.writelines(lines)


def update_man3_link():
    man3_path = os.path.join(environ.cache_dir, 'man3')

    if os.path.lexists(man3_path):
        if os.path.islink(man3_path):
            if os.readlink(man3_path) == environ.config.Source:
                return
            else:
                os.unlink(man3_path)
        else:
            raise RuntimeError("Can't create link since `%s' already exists" %
                               man3_path)
    try:
        os.makedirs(os.path.join(environ.cache_dir, environ.config.Source))
    except Exception:
        pass

    os.symlink(environ.config.Source, man3_path)


def get_width():
    """Get terminal width"""
    # Get terminal size
    columns, lines = shutil.get_terminal_size()
    width = min(columns * 39 // 40, columns - 2)
    return width


def groff2man(data):
    """Read groff-formatted text and output man pages."""
    width = get_width()

    cmd = 'groff -t -Tascii -m man -rLL=%dn -rLT=%dn' % (width, width)
    handle = subprocess.Popen(
        cmd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    man_text, stderr = handle.communicate(data)
    return man_text


def html2man(data, formatter):
    """Convert HTML text from cplusplus.com to man pages."""
    groff_text = formatter(data)
    man_text = groff2man(groff_text)
    return man_text


def fixupHTML(data):
    return str(bs4.BeautifulSoup(data, "html5lib"))

def urlopen(url, *args, **kwargs):
    """A wrapper around urllib.request.urlopen() which adds custom headers"""
    if isinstance(url, urllib.request.Request):
        req = url
    else:
        req = urllib.request.Request(url)
    req.add_header('User-Agent', _USER_AGENT)
    return urllib.request.urlopen(req, *args, **kwargs)

def build_opener(*args, **kwargs):
    """A wrapper around urllib.request.build_opener() which adds custom headers"""
    opener = urllib.request.build_opener(*args, **kwargs)
    opener.addheaders = [('User-Agent', _USER_AGENT)]
    return opener



# -*- coding: utf-8 -*-
#
# formatter.py - format html from cplusplus.com to groff syntax
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import datetime
import re

from cppman.formatter.tableparser import parse_table
from cppman.util import fixupHTML, html2man, urlopen

# Format replacement RE list
# The '.SE' pseudo macro is described in the function: html2groff
pre_rps = [
    # Snippet, ugly hack: we don't want to treat code listing as table
    (r'<table class="snippet">(.*?)</table>',
     r'\n.in +2n\n\1\n.in\n.sp\n', re.S),
]

rps = [
    # Header, Name
    (r'\s*<div id="I_type"[^>]*>(.*?)\s*</div>\s*'
     r'<div id="I_file"[^>]*>(.*?)</div>\s*'
     r'<h1>(.*?)</h1>\s*<div class="C_prototype"[^>]*>'
     r'(.*?)</div>\s*<div id="I_description"[^>]*>(.*?)</div>',
     r'.TH "\3" 3 "%s" "cplusplus.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n\3 - \5\n'
     r'\n.SE\n.SH "TYPE"\n\1\n'
     r'\n.SE\n.SH "SYNOPSIS"\n#include \2\n.sp\n\4\n'
     r'\n.SE\n.SH "DESCRIPTION"\n' % datetime.date.today(), re.S),
    (r'\s*<div id="I_type"[^>]*>(.*?)\s*</div>\s*'
     r'<div id="I_file"[^>]*>(.*?)</div>\s*'
     r'<h1>(.*?)</h1>\s*'
     r'<div id="I_description"[^>]*>(.*?)</div>',
     r'.TH "\3" 3 "%s" "cplusplus.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n\3 - \4\n'
     r'\n.SE\n.SH "TYPE"\n\1\n'
     r'\n.SE\n.SH "SYNOPSIS"\n#include \2\n.sp\n'
     r'\n.SE\n.SH "DESCRIPTION"\n' % datetime.date.today(), re.S),
    (r'\s*<div id="I_type"[^>]*>(.*?)\s*</div>\s*<h1>(.*?)</h1>\s*'
     r'<div id="I_description"[^>]*>(.*?)</div>',
     r'.TH "\2" 3 "%s" "cplusplus.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n\2 - \3\n'
     r'\n.SE\n.SH "TYPE"\n\1\n'
     r'\n.SE\n.SH "DESCRIPTION"\n' % datetime.date.today(), re.S),
    (r'\s*<div id="I_type"[^>]*>(.*?)\s*</div>\s*<h1>(.*?)</h1>\s*'
     r'<div id="I_file"[^>]*>(.*?)</div>\s*<div id="I_description"[^>]*>'
     '(.*?)</div>',
     r'.TH "\2" 3 "%s" "cplusplus.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n\2 - \4\n'
     r'\n.SE\n.SH "TYPE"\n\1\n'
     r'\n.SE\n.SH "DESCRIPTION"\n' % datetime.date.today(), re.S),
    (r'\s*<div id="I_type"[^>]*>(.*?)\s*</div>\s*<h1>(.*?)</h1>\s*'
     r'<div class="C_prototype"[^>]*>(.*?)</div>\s*'
     r'<div id="I_description"[^>]*>(.*?)</div>',
     r'.TH "\2" 3 "%s" "cplusplus.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n\2 - \4\n'
     r'\n.SE\n.SH "TYPE"\n\1\n'
     r'\n.SE\n.SH "SYNOPSIS"\n\3\n'
     r'\n.SE\n.SH "DESCRIPTION"\n' % datetime.date.today(), re.S),
    (r'<span alt="[^"]*?" class="C_ico cpp11warning"[^>]*>',
     r' [C++11]', re.S),
    # Remove empty #include
    (r'#include \n.sp\n', r'', 0),
    # Remove empty sections
    (r'\n.SH (.+?)\n+.SE', r'', 0),
    # Section headers
    (r'.*<h3>(.+?)</h3>', r'\n.SE\n.SH "\1"\n', 0),
    # 'ul' tag
    (r'<ul>', r'\n.RS 2\n', 0),
    (r'</ul>', r'\n.RE\n.sp\n', 0),
    # 'li' tag
    (r'<li>\s*(.+?)</li>', r'\n.IP \[bu] 3\n\1\n', re.S),
    # 'pre' tag
    (r'<pre[^>]*>(.+?)</pre\s*>', r'\n.nf\n\1\n.fi\n', re.S),
    # Subsections
    (r'<b>(.+?)</b>:<br/>', r'.SS \1\n', 0),
    # Member functions / See Also table
    # Without C++11 tag
    (r'<dl class="links"><dt><a href="[^"]*"><b>([^ ]+?)</b></a></dt><dd>'
     r'([^<]*?)<span class="typ">\s*\(([^<]*?)\n?\)</span></dd></dl>',
     r'\n.IP "\1 (3)"\n\2 (\3)\n', re.S),
    # With C++11 tag
    (r'<dl class="links"><dt><a href="[^"]*"><b>([^ ]+?) <b class="C_cpp11" '
     r'title="(.+?)"></b></b></a></dt><dd>'
     r'([^<]*?)<span class="typ">\s*\((.*?)\n?\)</span></dd></dl>',
     r'\n.IP "\1 (3) [\2]"\n\3 (\4)\n', re.S),
    # Footer
    (r'<div id="CH_bb">.*$',
     r'\n.SE\n.SH "REFERENCE"\n'
     r'cplusplus.com, 2000-2015 - All rights reserved.', re.S),
    # C++ version tag
    (r'<div.+?title="(C\+\+..)"[^>]*>', r'.sp\n\1\n', 0),
    # 'br' tag
    (r'<br/>', r'\n.br\n', 0),
    (r'\n.br\n.br\n', r'\n.sp\n', 0),
    # 'dd' 'dt' tag
    (r'<dt>(.+?)</dt>\s*<dd>(.+?)</dd>', r'.IP "\1"\n\2\n', re.S),
    # Bold
    (r'<strong>(.+?)</strong>', r'\n.B \1\n', 0),
    # Remove row number in EXAMPLE
    (r'<td class="rownum">.*?</td>', r'', re.S),
    # Any other tags
    (r'<script[^>]*>[^<]*</script>', r'', 0),
    (r'<.*?>', r'', re.S),
    # Misc
    (r'&lt;', r'<', 0),
    (r'&gt;', r'>', 0),
    (r'&quot;', r'"', 0),
    (r'&amp;', r'&', 0),
    (r'&nbsp;', r' ', 0),
    (r'\\([^\^nE])', r'\\\\\1', 0),
    (r'>/">', r'', 0),
    (r'/">', r'', 0),
    # Remove empty lines
    (r'\n\s*\n+', r'\n', 0),
    (r'\n\n+', r'\n', 0),
    # Preserve \n" in EXAMPLE
    (r'\\n', r'\\en', 0),
]


def escape_pre_section(table):
    """Escape <pre> section in table."""
    def replace_newline(g):
        return g.group(1).replace('\n', '\n.br\n')

    return re.sub('<pre.*?>(.*?)</pre>', replace_newline, table, flags=re.S)


def html2groff(data, name):
    """Convert HTML text from cplusplus.com to Groff-formatted text."""
    # Remove sidebar
    try:
        data = data[data.index('<div class="C_doc">'):]
    except ValueError:
        pass

    # Pre replace all
    for rp in pre_rps:
        data = re.compile(rp[0], rp[2]).sub(rp[1], data)

    for table in re.findall(r'<table.*?>.*?</table>', data, re.S):
        tbl = parse_table(escape_pre_section(table))
        # Escape column with '.' as prefix
        tbl = re.compile(r'T{\n(\..*?)\nT}', re.S).sub(r'T{\n\\E \1\nT}', tbl)
        data = data.replace(table, tbl)

    # Replace all
    for rp in rps:
        data = re.compile(rp[0], rp[2]).sub(rp[1], data)

    # Upper case all section headers
    for st in re.findall(r'.SH .*\n', data):
        data = data.replace(st, st.upper())

    # Add tags to member/inherited member functions
    # e.g. insert -> vector::insert
    #
    # .SE is a pseudo macro I created which means 'SECTION END'
    # The reason I use it is because I need a marker to know where section
    # ends.
    # re.findall find patterns which does not overlap, which means if I do
    # this: secs = re.findall(r'\n\.SH "(.+?)"(.+?)\.SH', data, re.S)
    # re.findall will skip the later .SH tag and thus skip the later section.
    # To fix this, '.SE' is used to mark the end of the section so the next
    # '.SH' can be find by re.findall

    page_type = re.search(r'\n\.SH "TYPE"\n(.+?)\n', data)
    if page_type and 'class' in page_type.group(1):
        class_name = re.search(
            r'\n\.SH "NAME"\n(?:.*::)?(.+?) ', data).group(1)

        secs = re.findall(r'\n\.SH "(.+?)"(.+?)\.SE', data, re.S)

        for sec, content in secs:
            # Member functions
            if ('MEMBER' in sec and
                'NON-MEMBER' not in sec and
                'INHERITED' not in sec and
                    sec != 'MEMBER TYPES'):
                content2 = re.sub(r'\n\.IP "([^:]+?)"', r'\n.IP "%s::\1"'
                                  % class_name, content)
                # Replace (constructor) (destructor)
                content2 = re.sub(r'\(constructor\)', r'%s' % class_name,
                                  content2)
                content2 = re.sub(r'\(destructor\)', r'~%s' % class_name,
                                  content2)
                data = data.replace(content, content2)
            # Inherited member functions
            elif 'MEMBER' in sec and 'INHERITED' in sec:
                inherit = re.search(r'.+?INHERITED FROM (.+)',
                                    sec).group(1).lower()
                content2 = re.sub(r'\n\.IP "(.+)"', r'\n.IP "%s::\1"'
                                  % inherit, content)
                data = data.replace(content, content2)

    # Remove pseudo macro '.SE'
    data = data.replace('\n.SE', '')

    return data


def func_test():
    """Test if there is major format changes in cplusplus.com"""
    ifs = urlopen('http://www.cplusplus.com/printf')
    result = html2groff(fixupHTML(ifs.read()), 'printf')
    assert '.SH "NAME"' in result
    assert '.SH "TYPE"' in result
    assert '.SH "DESCRIPTION"' in result


def test():
    """Simple Text"""
    ifs = urlopen('http://www.cplusplus.com/vector')
    print(html2groff(fixupHTML(ifs.read()), 'std::vector'), end=' ')
    # with open('test.html') as ifs:
    #    print html2groff(fixupHTML(ifs.read()), 'std::vector'),


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# environ.py
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os

from cppman import get_lib_path
from cppman.config import Config

HOME = os.environ["HOME"]

XDG_CACHE_HOME = os.getenv("XDG_CACHE_HOME", os.path.join(HOME, ".cache"))
XDG_CONFIG_HOME = os.getenv("XDG_CONFIG_HOME", os.path.join(HOME, ".config"))

cache_dir = os.path.join(XDG_CACHE_HOME, 'cppman')
manindex_dir = os.path.join(cache_dir, 'manindex')
config_dir = os.path.join(XDG_CONFIG_HOME, 'cppman')
config_file = os.path.join(config_dir, 'cppman.cfg')

config = Config(config_file)

try:
    os.makedirs(cache_dir)
    os.makedirs(manindex_dir)
    os.makedirs(config_dir)
    update_man3_link()
except:
    pass

index_db_re = os.path.join(cache_dir, 'index.db')

index_db = index_db_re if os.path.exists(index_db_re) \
    else get_lib_path('index.db')

pager = config.Pager
pager_config = get_lib_path('cppman.vim')
pager_script = get_lib_path('pager.sh')

source = config.Source
if source not in config.SOURCES:
    source = config.SOURCES[0]
    config.Source = source



# -*- coding: utf-8 -*-
#
# main.py
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import collections
import gzip
import html
import importlib
import os
import os.path
import re
import shutil
import sqlite3
import subprocess
import sys

from bs4 import BeautifulSoup
from cppman import environ, util
from cppman.crawler import Crawler
from urllib.parse import urlparse, unquote



def _sort_crawl(entry):
    """ Sorting entries for putting '(1)' indexes behind keyword

        1. keywords that have 'std::' in them have highest priority
        2. priority if 'std::' is inside their name
        3. sorting by keyword
        4. sorting by name
    """
    id, title, keyword, count = entry
    hasStd1 = keyword.find("std::")
    if hasStd1 == -1:
        hasStd1 = 1
    else:
        hasStd1 = 0

    hasStd2 = title.find("std::")
    if hasStd2 == -1:
        hasStd2 = 1
    else:
        hasStd2 = 0

    return (hasStd1, hasStd2, keyword, title)


def _sort_search(entry, pattern):
    """ Sort results
        0. exact match goes first
        1. sort by 'std::' (an entry with `std::` goes before an entry without)
        2. sort by which position the keyword appears
    """

    title, keyword, url = entry

    if keyword == pattern:
        # Exact match - lowest key value
        return (-1, -1, 0, keyword)

    hasStd1 = keyword.find("std::")
    if hasStd1 == -1:
        hasStd1 = 1
    else:
        hasStd1 = 0

    hasStd2 = title.find("std::")
    if hasStd2 == -1:
        hasStd2 = 1
    else:
        hasStd2 = 0

    return (hasStd1, hasStd2, keyword.find(pattern), keyword)

# Return the longest prefix of all list elements.
def _commonprefix(s1, s2):
    """" Given two strings, returns the longest common leading prefix """

    if len(s1) > len(s2):
        s1, s2 = s2, s1;
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1

def _removeprefix(string, prefix):
    if prefix and string.startswith(prefix):
        return string[len(prefix):]
    return string

def _removesuffix(string, suffix):
    if suffix and string.endswith(suffix):
        return string[:-len(suffix)]
    return string

class Cppman(Crawler):
    """ Manage cpp man pages, indexes. """

    def __init__(self, forced=False, force_columns=-1):
        Crawler.__init__(self)
        self.forced = forced
        self.success_count = None
        self.failure_count = None
        self.force_columns = force_columns

    def rebuild_index(self):
        """ Rebuild index database from cplusplus.com and cppreference.com. """

        self.db_conn = sqlite3.connect(environ.index_db_re)
        self.db_cursor = self.db_conn.cursor()
        try:
            self.add_url_filter(r'\.(jpg|jpeg|gif|png|js|css|swf|svg)$')
            self.set_follow_mode(Crawler.F_SAME_PATH)

            sources = [('cplusplus.com', 'https://cplusplus.com/reference/', None),
                       ('cppreference.com', 'https://en.cppreference.com/w/cpp', '/w/cpp')]

            for table, url, path in sources:
                """ Drop and recreate tables. """
                self.db_cursor.execute(
                    'DROP TABLE IF EXISTS "%s"'
                    % table)

                self.db_cursor.execute(
                    'DROP TABLE IF EXISTS "%s_keywords"'
                    % table)

                self.db_cursor.execute(
                    'CREATE TABLE "%s" ('
                    'id INTEGER NOT NULL PRIMARY KEY, '
                    'title VARCHAR(255) NOT NULL UNIQUE, '
                    'url VARCHAR(255) NOT NULL UNIQUE'
                    ')' % table)

                self.db_cursor.execute(
                    'CREATE TABLE "%s_keywords" ('
                    'id INTEGER NOT NULL, '
                    'keyword VARCHAR(255), '
                    'FOREIGN KEY(id) REFERENCES "%s"(id)'
                    ')' % (table, table))

                """ Crawl and insert all entries. """
                self.results = collections.defaultdict(list)
                self.crawl(url)
                results = self._results_with_unique_title()

                for title in results:
                    """ 1. insert title """
                    self.db_cursor.execute(
                        'INSERT INTO "%s" (title, url) VALUES (?, ?)'
                        % table, (title, results[title]["url"]))

                    lastRow = self.db_cursor.execute(
                        'SELECT last_insert_rowid()').fetchall()[0][0]

                    """ 2. insert all keywords """
                    for k in results[title]["keywords"]:
                        self.db_cursor.execute(
                            'INSERT INTO "%s_keywords" (id, keyword) '
                            'VALUES (?, ?)'
                            % table, (lastRow, k))

                """ 3. add all aliases """
                for title in results:
                    for (k, a) in results[title]["aliases"]:
                        """ search for combinations of words

                            e.g. std::basic_string::append
                        """
                        sql_results = self.db_cursor.execute(
                            'SELECT id, keyword FROM "%s_keywords" '
                            'WHERE keyword LIKE "%%::%s::%%" '
                            'OR keyword LIKE "%s::%%" '
                            'OR keyword LIKE "%s" '
                            'OR keyword LIKE "%s %%" '
                            'OR keyword LIKE "%s)%%" '
                            'OR keyword LIKE "%s,%%"'
                            % (table, k, k, k, k, k, k)).fetchall()

                        for id, keyword in sql_results:
                            keyword = re.sub(re.escape("%s" % k), "%s" % a, keyword, flags=re.IGNORECASE)

                            self.db_cursor.execute(
                                'INSERT INTO "%s_keywords" (id, keyword) '
                                'VALUES (?, ?)'
                                % table, (id, keyword))

                self.db_conn.commit()

                """ remove duplicate keywords that link the same page """
                self.db_cursor.execute(
                    'DELETE FROM "%s_keywords" WHERE rowid NOT IN ('
                    'SELECT min(rowid) FROM "%s_keywords" '
                    'GROUP BY id, keyword '
                    ')' % (table, table)).fetchall()

                """ give duplicate keywords with different links entry numbers """
                results = self.db_cursor.execute(
                    'SELECT t3.id, t3.title, t2.keyword, t1.count '
                    'FROM ('
                    '      SELECT keyword, COUNT(*) AS count FROM "%s_keywords" '
                    '      GROUP BY keyword HAVING count > 1) AS t1 '
                    'JOIN "%s_keywords" AS t2 '
                    'JOIN "%s" AS t3 '
                    'WHERE t1.keyword = t2.keyword AND t3.id = t2.id '
                    'ORDER BY t2.keyword, t3.title'
                    % (table, table, table)).fetchall()

                keywords = {}
                results = sorted(results, key=_sort_crawl)
                for id, title, keyword, count in results:
                    if not keyword in keywords:
                        keywords[keyword] = 0
                    keywords[keyword] += 1
                    new_keyword = "%s (%s)" % (keyword, keywords[keyword])
                    self.db_cursor.execute(
                        'UPDATE "%s_keywords" SET keyword=? WHERE '
                        'id=? AND keyword=?'
                        % table, (new_keyword, id, keyword))

                self.db_conn.commit()

        except KeyboardInterrupt:
            os.remove(environ.index_db_re)
            raise KeyboardInterrupt
        finally:
            self.db_conn.close()

    def process_document(self, url, content, depth):
        """callback to insert index"""
        print("Indexing '%s' (depth %s)..." % (url, depth))
        name = self._extract_name(content).replace('\n', '')
        keywords = self._extract_keywords(content)

        entry = {'url': url, 'keywords': set(), 'aliases': set()}
        self.results[name].append(entry)

        for n in self._parse_title(name):
            """ add as keyword """
            entry["keywords"].add(n)

            """ add as keyword without std:: """
            if n.find("std::") != -1:
                entry["keywords"].add(n.replace('std::', ''))

            """ add with all keywords variations """
            for k in keywords:
                """ add std:: to typedef if original type is in std namespace """
                if n.find("std::") != -1 and k.find("std::") == -1:
                    k = "std::" + k;

                entry["aliases"].add((n, k))
                prefix = _commonprefix(n, k)

                if len(prefix) > 2 and prefix[-2:] == "::":
                    """ Create names and keyword without prefixes """
                    new_name = n[len(prefix):]
                    new_key  = k[len(prefix):]
                    entry["aliases"].add((new_name, new_key))

                if k.find("std::") != -1:
                    entry["aliases"].add(
                        (n, k.replace('std::', '')))

        return True

    def _results_with_unique_title(self):
        """process crawling results and return title -> entry dictionary;
           add part of the path to entries having the same title
        """
        results = dict()
        for title, entries in self.results.items():
            if len(entries) == 1:
                results[title] = entries[0]
            else:
                paths = [_removesuffix(urlparse(entry['url'])[2], '/') for entry in entries]
                prefix = os.path.commonpath(paths)
                if prefix:
                    prefix += '/'
                suffix = '/' + os.path.basename(paths[0])
                for path in paths:
                    if not path.endswith(suffix):
                        suffix = ''
                        break
                for index, entry in enumerate(entries):
                    path = _removeprefix(paths[index], prefix)
                    path = _removesuffix(path, suffix)
                    results["{} ({})".format(title, unquote(path))] = entry
        return results

    def _extract_name(self, data):
        """Extract man page name from web page."""
        name = re.search('<[hH]1[^>]*>(.+?)</[hH]1>', data, re.DOTALL).group(1)
        name = re.sub(r'<([^>]+)>', r'', name)
        name = re.sub(r'&gt;', r'>', name)
        name = re.sub(r'&lt;', r'<', name)
        return html.unescape(name)

    def _parse_expression(self, expr):
        """
            split expression into prefix and expression
            tested with
            ```
            operator==
            !=
            std::rel_ops::operator!=
            std::atomic::operator=
            std::array::operator[]
            std::function::operator()
            std::vector::at
            std::relational operators
            std::vector::begin
            std::abs(float)
            std::fabs()
            ```
        """
        m = re.match(r'^(.*?(?:::)?(?:operator)?)((?:::[^:]*|[^:]*)?)$', expr)
        prefix = m.group(1)
        tail = m.group(2)
        return [prefix, tail]

    def _parse_title(self, title):
        """
             split of the last parenthesis  operator==,!=,<,<=(std::vector)
             tested with
            ```
            operator==,!=,<,<=,>,>=(std::vector)
            operator==,!=,<,<=,>,>=(std::vector)
            operator==,!=,<,<=,>,>=
            operator==,!=,<,<=,>,>=
            std::rel_ops::operator!=,>,<=,>=
            std::atomic::operator=
            std::array::operator[]
            std::function::operator()
            std::vector::at
            std::relational operators (vector)
            std::vector::begin, std::vector::cbegin
            std::abs(float), std::fabs
            std::unordered_set::begin(size_type), std::unordered_set::cbegin(size_type)
            ```
        """
        """ remove all template stuff """
        title = re.sub(r" ?<[^>]+>", "", title)

        m = re.match(
            r'^\s*((?:\(size_type\)|(?:.|\(\))*?)*)((?:\([^)]+\))?)\s*$', title)

        postfix = m.group(2)

        t_names = m.group(1).split(',')
        t_names = [n.strip() for n in t_names]
        prefix = self._parse_expression(t_names[0])[0]
        names = []
        for n in t_names:
            r = self._parse_expression(n)
            if prefix == r[0]:
                names.append(n + postfix)
            else:
                names.append(prefix + r[1] + postfix)
        return names

    def _extract_keywords(self, text):
        """
            extract aliases like std::string, template specializations like std::atomic_bool
            and helper functions like std::is_same_v
        """
        soup = BeautifulSoup(text, "lxml")
        names = []

        # search for typedef list
        for x in soup.find_all('table'):
            # just searching for "Type" is not enough, see std::is_same
            p = x.find_previous_sibling('h3')
            if p:
                if p.get_text().strip() == "Member types":
                    continue

            typedefTable = False
            for tr in x.find_all('tr'):
                tds = tr.find_all('td')
                if len(tds) == 2:
                    if re.match(r"\s*Type\s*", tds[0].get_text()):
                        typedefTable = True
                    elif typedefTable:
                        res = re.search(r'^\s*(\S*)\s+.*$', tds[0].get_text())
                        if res and res.group(1):
                            names.append(res.group(1))
                    elif not typedefTable:
                        break
            if typedefTable:
                break

        # search for "Helper variable template" list
        for x in soup.find_all('h3'):
            variableTemplateHeader = False
            if x.find('span', id="Helper_variable_template"):
                e = x.find_next_sibling()
                while e.name == "":
                    e = e.find_next_sibling()
                if e.name == "table":
                    for tr in e.find_all('tr'):
                        text = re.sub('\n', ' ', tr.get_text())
                        res = re.search(r'^.* (\S+)\s*=.*$', text)
                        if res:
                            names.append(res.group(1))
        # search for "Helper types" list
        for x in soup.find_all('h3'):
            variableTemplateHeader = False
            if x.find('span', id="Helper_types"):
                e = x.find_next_sibling()
                while e.name == "":
                    e = e.find_next_sibling()
                if e.name == "table":
                    for tr in e.find_all('tr'):
                        text = re.sub('\n', ' ', tr.get_text())
                        res = re.search(r'^.* (\S+)\s*=.*$', text)
                        if res:
                            names.append(res.group(1))
        return [html.unescape(n) for n in names]

    def cache_all(self):
        """Cache all available man pages"""

        respond = input(
            'By default, cppman fetches pages on-the-fly if corresponding '
            'page is not found in the cache. The "cache-all" option is only '
            'useful if you want to view man pages offline. '
            'Caching all contents will take several minutes, '
            'do you want to continue [y/N]? ')
        if not (respond and 'yes'.startswith(respond.lower())):
            raise KeyboardInterrupt

        try:
            os.makedirs(environ.cache_dir)
        except:
            pass

        self.success_count = 0
        self.failure_count = 0

        if not os.path.exists(environ.index_db):
            raise RuntimeError("can't find index.db")

        conn = sqlite3.connect(environ.index_db)
        cursor = conn.cursor()

        source = environ.config.source
        print('Caching manpages from %s ...' % source)
        data = cursor.execute('SELECT title, url FROM "%s"' % source).fetchall()

        for name, url in data:
            print('Caching %s ...' % name)
            retries = 3
            while retries > 0:
                try:
                    self.cache_man_page(source, url, name)
                except Exception:
                    print('Retrying ...')
                    retries -= 1
                else:
                    self.success_count += 1
                    break
            else:
                print('Error caching %s ...' % name)
                self.failure_count += 1

        conn.close()

        print('\n%d manual pages cached successfully.' % self.success_count)
        print('%d manual pages failed to cache.' % self.failure_count)
        self.update_mandb(False)

    def cache_man_page(self, source, url, name):
        """callback to cache new man page"""
        # Skip if already exists, override if forced flag is true
        outname = self.get_page_path(source, name)
        if os.path.exists(outname) and not self.forced:
            return

        try:
            os.makedirs(os.path.join(environ.cache_dir, source))
        except OSError:
            pass

        # There are often some errors in the HTML, for example: missing closing
        # tag. We use fixupHTML to fix this.
        data = util.fixupHTML(util.urlopen(url).read())

        formatter = importlib.import_module(
            'cppman.formatter.%s' % source[:-4])
        groff_text = formatter.html2groff(data, name)

        with gzip.open(outname, 'w') as f:
            f.write(groff_text.encode('utf-8'))

    def clear_cache(self):
        """Clear all cache in man"""
        shutil.rmtree(environ.cache_dir)

    def _fetch_page_by_keyword(self, keyword):
        """ fetches result for a keyword """
        return self.cursor.execute(
            'SELECT t1.title, t2.keyword, t1.url '
            'FROM "%s" AS t1 '
            'JOIN "%s_keywords" AS t2 '
            'WHERE t1.id = t2.id AND t2.keyword '
            'LIKE ? ORDER BY t2.keyword'
            % (self.source, self.source), ['%%%s%%' % keyword]).fetchall()

    def _search_keyword(self, pattern):
        """ multiple fetches for each pattern """
        if not os.path.exists(environ.index_db):
            raise RuntimeError("can't find index.db")

        conn = sqlite3.connect(environ.index_db)
        self.cursor = conn.cursor()
        self.source = environ.source

        self.cursor.execute('PRAGMA case_sensitive_like=ON')
        results = self._fetch_page_by_keyword("%s" % pattern)
        results.extend(self._fetch_page_by_keyword("%s %%" % pattern))
        results.extend(self._fetch_page_by_keyword("%% %s" % pattern))
        results.extend(self._fetch_page_by_keyword("%% %s %%" % pattern))

        results.extend(self._fetch_page_by_keyword("%s%%" % pattern))
        if len(results) == 0:
            results = self._fetch_page_by_keyword("%%%s%%" % pattern)

        conn.close()
        return sorted(list(set(results)), key=lambda e: _sort_search(e, pattern))

    def man(self, pattern):
        """Call viewer.sh to view man page"""
        results = self._search_keyword(pattern)
        if len(results) == 0:
            raise RuntimeError('No manual entry for %s ' % pattern)

        page_name, keyword, url = results[0]

        try:
            avail = os.listdir(os.path.join(environ.cache_dir, environ.source))
        except OSError:
            avail = []

        page_filename = self.get_normalized_page_name(page_name)
        if self.forced or page_filename + '.3.gz' not in avail:
            self.cache_man_page(environ.source, url, page_name)

        pager_type = environ.pager if sys.stdout.isatty() else 'pipe'

        # Call viewer
        columns = (util.get_width() if self.force_columns == -1 else
                   self.force_columns)
        pid = os.fork()
        if pid == 0:
            os.execl('/bin/sh', '/bin/sh', environ.pager_script, pager_type,
                     self.get_page_path(environ.source, page_name),
                     str(columns), environ.pager_config, pattern)
        return pid

    def find(self, pattern):
        """Find pages in database."""

        results = self._search_keyword(pattern)

        pat = re.compile(r'(.*?)(%s)(.*?)( \(.*\))?$' %
                         re.escape(pattern), re.I)

        if results:
            for name, keyword, url in results:

                if os.isatty(sys.stdout.fileno()):
                    keyword = pat.sub(
                        r'\1\033[1;31m\2\033[0m\3\033[1;33m\4\033[0m', keyword)
                print("%s - %s" % (keyword, name))
        else:
            raise RuntimeError('%s: nothing appropriate.' % pattern)

    def fuzzy_find(self, pattern):
        """Find pages in database and present an interactive selection menu."""
        results = self._search_keyword(pattern)

        if not results:
            raise RuntimeError('%s: nothing appropriate.' % pattern)

        if len(results) == 1:
            return results[0][1]

        for i, (name, keyword, url) in enumerate(results, 1):
            print(f"{i}. {keyword} - {name}")

        while True:
            try:
                selection = input("\nPlease enter the selection: ")
                if not selection:
                    return None

                idx = int(selection) - 1
                if 0 <= idx < len(results):
                    return results[idx][1]
                print("Invalid selection. Please try again.")
            except ValueError:
                print("Please enter a valid number.")
            except KeyboardInterrupt:
                print("\nOperation cancelled.")
                return None

    def update_mandb(self, quiet=True):
        """Update mandb."""
        if not environ.config.UpdateManPath:
            return
        print('\nrunning mandb...')
        cmd = 'mandb %s' % (' -q' if quiet else '')
        subprocess.Popen(cmd, shell=True).wait()

    def get_normalized_page_name(self, name):
        return name.replace('/', '_')

    def get_page_path(self, source, name):
        name = self.get_normalized_page_name(name)
        return os.path.join(environ.cache_dir, source, name + '.3.gz')



# -*- coding: utf-8 -*-
#
# config.py
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import configparser
import os


class Config(object):
    PAGERS = ['vim', 'nvim', 'less', 'system']
    SOURCES = ['cplusplus.com', 'cppreference.com']

    DEFAULTS = {
        'Source': 'cppreference.com',
        'UpdateManPath': 'false',
        'Pager': 'less'
    }

    def __init__(self, configfile):
        self._configfile = configfile

        if not os.path.exists(configfile):
            self.set_default()
        else:
            self._config = configparser.RawConfigParser()
            self._config.read(self._configfile)

    def __getattr__(self, name):
        try:
            value = self._config.get('Settings', name)
        except configparser.NoOptionError:
            value = self.DEFAULTS[name]
            setattr(self, name, value)
            self._config.read(self._configfile)

        return self.parse_bool(value)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            self._config.set('Settings', name, value)
            self.save()
        self.__dict__[name] = self.parse_bool(value)

    def set_default(self):
        """Set config to default."""
        try:
            os.makedirs(os.path.dirname(self._configfile))
        except:
            pass

        self._config = configparser.RawConfigParser()
        self._config.add_section('Settings')

        for key, val in self.DEFAULTS.items():
            self._config.set('Settings', key, val)

        with open(self._configfile, 'w') as f:
            self._config.write(f)

    def save(self):
        """Store config back to file."""
        try:
            os.makedirs(os.path.dirname(self._configfile))
        except:
            pass

        with open(self._configfile, 'w') as f:
            self._config.write(f)

    def parse_bool(self, val):
        if type(val) == str:
            if val.lower() == 'true':
                return True
            elif val.lower() == 'false':
                return False
        return val



# -*- coding: utf-8 -*-
#
# formatter.py - format html from cplusplus.com to groff syntax
#
# Copyright (C) 2010 - 2015  Wei-Ning Huang (AZ) <aitjcize@gmail.com>
# All Rights reserved.
#
# This file is part of cppman.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import datetime
import re
import string
from functools import partial

from cppman.formatter.tableparser import parse_table
from cppman.util import fixupHTML, html2man, urlopen


def member_table_def(g):
    tbl = parse_table('<table>%s</table>' % str(g.group(3)))
    # Escape column with '.' as prefix
    tbl = re.compile(r'T{\n(\..*?)\nT}', re.S).sub(r'T{\n\\E \1\nT}', tbl)
    return '\n.IP "%s"\n%s\n%s\n' % (g.group(1), g.group(2), tbl)


def member_type_function(g):
    if g.group(1).find("<a href=") == -1:
        return ""
    head = re.sub(r'<.*?>', '', g.group(1)).strip()
    tail = ''

    spectag = re.search(r'^(.*?)(\[(?:static|virtual)\])(.*)$', head)
    if spectag:
        head = spectag.group(1).strip() + ' ' + spectag.group(3).strip()
        tail = ' ' + spectag.group(2)

    cppvertag = re.search(
        r'^(.*?)(\[(?:(?:since|until) )?C\+\+\d+\]\s*(,\s*)?)+$', head)
    if cppvertag:
        head = cppvertag.group(1).strip()
        tail = ' ' + cppvertag.group(2)

    if ',' in head:
        head = ', '.join([x.strip() + ' (3)' for x in head.split(',')])
    else:
        head = head.strip() + ' (3)'

    full = (head + tail).replace('"', '\\(dq')
    return '\n.IP "%s"\n%s\n' % (full, g.group(2))


NAV_BAR_END = '<div class="t-navbar-sep">.?</div></div>'

# Format replacement RE list
# The '.SE' pseudo macro is described in the function: html2groff
rps = [
    # Workaround: remove <p> in t-dcl
    (r'<tr class="t-dcl">(.*?)</tr>',
     lambda g: re.sub('<p/?>', '', g.group(1)), re.S),
    # Header, Name
    (r'<h1.*?>(.*?)</h1>',
     r'\n.TH "{{name}}" 3 "%s" "cppreference.com" "C++ Programmer\'s Manual"\n'
     r'\n.SH "NAME"\n{{name}} {{shortdesc}}\n.SE\n' % datetime.date.today(),
     re.S),
    # Defined in header
    (r'<div class="t-navbar"[^>]*>.*?' + NAV_BAR_END + r'.*?'
     r'Defined in header <code>(.*?)</code>(.*?)<tr class="t-dcl-sep">',
     r'\n.SH "SYNOPSIS"\n#include \1\n.sp\n'
     r'.nf\n\2\n.fi\n.SE\n'
     r'\n.SH "DESCRIPTION"\n', re.S),
    (r'<div class="t-navbar"[^>]*>.*?' + NAV_BAR_END +
     r'(.*?)<tr class="t-dcl-sep">',
     r'\n.SH "SYNOPSIS"\n.nf\n\1\n.fi\n.SE\n'
     r'\n.SH "DESCRIPTION"\n', re.S),
    # <unordered_map>
    (r'<div class="t-navbar"[^>]*>.*?' + NAV_BAR_END +
     r'(.*?)<table class="t-dsc-begin">',
     r'\n.SH "DESCRIPTION"\n\1\n', re.S),
    # access specifiers
    (r'<div class="t-navbar"[^>]*>.*?' + NAV_BAR_END +
     r'(.*?)<h3',
     r'\n.SH "DESCRIPTION"\n\1\n<h3', re.S),
    (r'<td>\s*\([0-9]+\)\s*</td>', r'', 0),
    # Section headers
    (r'<div class="t-inherited">.*?<h2>.*?Inherited from\s*(.*?)\s*</h2>',
     r'\n.SE\n.IEND\n.IBEGIN \1\n', re.S),
    # Remove tags
    (r'<span class="edit.*?">.*?</span> ?', r'', re.S),
    (r'&#91;edit&#93;', r'', re.S),
    (r'\[edit\]', r'', re.S),
    (r'<div id="siteSub">.*?</div>', r'', 0),
    (r'<div id="contentSub">.*?</div>', r'', 0),
    (r'<table class="toc" id="toc"[^>]*>.*?</table>', r'', re.S),
    (r'<h2[^>]*>.*?</h2>', r'', re.S),
    (r'<div class="coliru-btn coliru-btn-run-init">.*?</div>', r'', re.S),
    (r'<tr class="t-dsc-hitem">.*?</tr>', r'', re.S),
    # C++11/14/17/20
    (r'\(((?:since|until) C\+\+\d+)\)', r' [\1]', re.S),
    (r'\((C\+\+\d+)\)', r' [\1]', re.S),
    # Subsections
    (r'<h5[^>]*>\s*(.*)</h5>', r'\n.SS "\1"\n', 0),
    # Group t-lines
    (r'<span></span>', r'', re.S),
    (r'<span class="t-lines">(?:<span>.+?</span>.*)+</span>',
     lambda x: re.sub(r'\s*</span><span>\s*', r', ', x.group(0)), re.S),
    # Member type & function second col is group see basic_fstream for example
    (r'<tr class="t-dsc">\s*?<td>((?:(?!</td>).)*?)</td>\s*?'
     r'<td>((?:(?!</td>).)*?)<table[^>]*>((?:(?!</table>).)*?)</table>'
     r'(?:(?!</td>).)*?</td>\s*?</tr>',
     member_table_def, re.S),
    # Section headers
    (r'.*<h3>(.+?)</h3>', r'\n.SE\n.SH "\1"\n', 0),
    # Member type & function
    (r'<tr class="t-dsc">\n?<td>\s*(.*?)\n?</td>.*?<td>\s*(.*?)</td>.*?</tr>',
     member_type_function, re.S),
    # Parameters
    (r'<tr class="t-par">.*?<td>\s*(.*?)\n?</td>.*?<td>.*?</td>.*?'
     r'<td>\s*(.*?)</td>.*?</tr>',
     r'\n.IP "\1"\n\2\n', re.S),
    # 'ul' tag
    (r'<ul>', r'\n.RS 2\n', 0),
    (r'</ul>', r'\n.RE\n.sp\n', 0),
    # 'li' tag
    (r'<li>\s*(.+?)</li>', r'\n.IP \[bu] 3\n\1\n', re.S),
    # 'pre' tag
    (r'<pre[^>]*>(.+?)</pre\s*>', r'\n.in +2n\n.nf\n\1\n.fi\n.in\n', re.S),
    # Footer
    (r'<div class="printfooter">',
     r'\n.SE\n.IEND\n.SH "REFERENCE"\n'
     r'cppreference.com, 2015 - All rights reserved.', re.S),
    # C++ version tag
    (r'<div title="(C\+\+..)"[^>]*>', r'.sp\n\1\n', 0),
    # Output
    (r'<p>Output:\n?</p>', r'\n.sp\nOutput:\n', re.S),
    # Paragraph
    (r'<p>(.*?)</p>', r'\n\1\n.sp\n', re.S),
    (r'<div class="t-li1">(.*?)</div>', r'\n\1\n.sp\n', re.S),
    (r'<div class="t-li2">(.*?)</div>',
     r'\n.RS\n\1\n.RE\n.sp\n', re.S),
    # 'br' tag
    (r'<br/>', r'\n.br\n', 0),
    (r'\n.br\n.br\n', r'\n.sp\n', 0),
    # 'dd' 'dt' tag
    (r'<dt>(.+?)</dt>\s*<dd>(.+?)</dd>', r'\n.IP "\1"\n\2\n', re.S),
    # Bold
    (r'<strong>(.+?)</strong>', r'\n.B \1\n', 0),
    # Any other tags
    (r'<script[^>]*>[^<]*</script>', r'', 0),
    (r'<.*?>', r'', re.S),
    # Escape
    (r'^#', r'\#', 0),
    (r'&#160;', ' ', 0),
    (r'&#(\d+);', lambda g: chr(int(g.group(1))), 0),
    # Misc
    (r'&lt;', r'<', 0),
    (r'&gt;', r'>', 0),
    (r'&quot;', r'"', 0),
    (r'&amp;', r'&', 0),
    (r'&nbsp;', r' ', 0),
    (r'\\([^\^nE])', r'\\\\\1', 0),
    (r'>/">', r'', 0),
    (r'/">', r'', 0),
    # Remove empty sections
    (r'\n.SH (.+?)\n+.SE', r'', 0),
    # Remove empty lines
    (r'\n\s*\n+', r'\n', 0),
    (r'\n\n+', r'\n', 0),
    # Preserve \n" in EXAMPLE
    (r'\\n', r'\\en', 0),
    # Remove leading whitespace
    (r'^\s+', r'', re.S),
    # Trailing white-spaces
    (r'\s+\n', r'\n', re.S),
    # Remove extra whitespace and newline in .SH/SS/IP section
    (r'.(SH|SS|IP) "\s*(.*?)\s*\n?"', r'.\1 "\2"', 0),
    # Remove extra whitespace before .IP bullet
    (r'(.IP \\\\\[bu\] 3)\n\s*(.*?)\n', r'\1\n\2\n', 0),
    # Remove extra '\n' before C++ version Tag (don't do it in table)
    (r'(?<!T{)\n\s*(\[(:?since|until) C\+\+\d+\])', r' \1', re.S)
]


def html2groff(data, name):
    """Convert HTML text from cppreference.com to Groff-formatted text."""
    # Remove header and footer
    try:
        data = data[data.index('<div id="cpp-content-base">'):]
        data = data[:data.index('<div class="printfooter">') + 25]
    except ValueError:
        pass

    # Remove non-printable characters
    data = ''.join([x for x in data if x in string.printable])

    for table in re.findall(
            r'<table class="(?:wikitable|dsctable)"[^>]*>.*?</table>',
            data, re.S):
        tbl = parse_table(table)
        # Escape column with '.' as prefix
        tbl = re.compile(r'T{\n(\..*?)\nT}', re.S).sub(r'T{\n\\E \1\nT}', tbl)
        data = data.replace(table, tbl)

    # Pre replace all
    for rp in rps:
        data = re.compile(rp[0], rp[2]).sub(rp[1], data)

    # Remove non-printable characters
    data = ''.join([x for x in data if x in string.printable])

    # Upper case all section headers
    for st in re.findall(r'.SH .*\n', data):
        data = data.replace(st, st.upper())

    # Add tags to member/inherited member functions
    # e.g. insert -> vector::insert
    #
    # .SE is a pseudo macro I created which means 'SECTION END'
    # The reason I use it is because I need a marker to know where section
    # ends.
    # re.findall find patterns which does not overlap, which means if I do
    # this: secs = re.findall(r'\n\.SH "(.+?)"(.+?)\.SH', data, re.S)
    # re.findall will skip the later .SH tag and thus skip the later section.
    # To fix this, '.SE' is used to mark the end of the section so the next
    # '.SH' can be find by re.findall

    try:
        idx = data.index('.IEND')
    except ValueError:
        idx = None

    def add_header_multi(prefix, g):
        if ',' in g.group(1):
            res = ', '.join(['%s::%s' % (prefix, x.strip())
                             for x in g.group(1).split(',')])
        else:
            res = '%s::%s' % (prefix, g.group(1))

        return '\n.IP "%s"' % res

    if idx:
        class_name = name
        if class_name.startswith('std::'):
            normalized_class_name = class_name[len('std::'):]
        else:
            normalized_class_name = class_name
        class_member_content = data[:idx]
        secs = re.findall(r'\.SH "(.+?)"(.+?)\.SE', class_member_content, re.S)

        for sec, content in secs:
            # Member functions
            if (('MEMBER' in sec and
                'NON-MEMBER' not in sec and
                'INHERITED' not in sec and
                'MEMBER TYPES' != sec) or
                'CONSTANTS' == sec):
                content2 = re.sub(r'\n\.IP "([^:]+?)"',
                                  partial(add_header_multi, class_name),
                                  content)
                # Replace (constructor) (destructor)
                content2 = re.sub(r'\(constructor\)', r'%s' %
                                  normalized_class_name, content2)
                content2 = re.sub(r'\(destructor\)', r'~%s' %
                                  normalized_class_name, content2)
                data = data.replace(content, content2)

    blocks = re.findall(r'\.IBEGIN\s*(.+?)\s*\n(.+?)\.IEND', data, re.S)

    for inherited_class, content in blocks:
        content2 = re.sub(r'\.SH "(.+?)"', r'\n.SH "\1 INHERITED FROM %s"'
                          % inherited_class.upper(), content)
        data = data.replace(content, content2)

        secs = re.findall(r'\.SH "(.+?)"(.+?)\.SE', content, re.S)

        for sec, content in secs:
            # Inherited member functions
            if 'MEMBER' in sec and \
               sec != 'MEMBER TYPES':
                content2 = re.sub(r'\n\.IP "(.+)"',
                                  partial(add_header_multi, inherited_class),
                                  content)
                data = data.replace(content, content2)

    # Remove unneeded pseudo macro
    data = re.sub('(?:\n.SE|.IBEGIN.*?\n|\n.IEND)', '', data)

    # Replace all macros
    desc_re = re.search(r'.SH "DESCRIPTION"\n.*?([^\n\s].*?)\n', data)
    shortdesc = ''

    # not empty description
    if desc_re and not desc_re.group(1).startswith('.SH'):
        shortdesc = '- ' + desc_re.group(1)

    def dereference(g):
        d = dict(name=name, shortdesc=shortdesc)
        if g.group(1) in d:
            return d[g.group(1)]

    data = re.sub('{{(.*?)}}', dereference, data)

    return data


def func_test():
    """Test if there is major format changes in cplusplus.com"""
    ifs = urlopen('http://en.cppreference.com/w/cpp/container/vector')
    result = html2groff(fixupHTML(ifs.read()), 'std::vector')
    assert '.SH "NAME"' in result
    assert '.SH "SYNOPSIS"' in result
    assert '.SH "DESCRIPTION"' in result


def test():
    """Simple Text"""
    ifs = urlopen('http://en.cppreference.com/w/cpp/container/vector')
    print(html2groff(fixupHTML(ifs.read()), 'std::vector'), end=' ')
    # with open('test.html') as ifs:
    #    data = fixupHTML(ifs.read())
    #    print html2groff(data, 'std::vector'),


if __name__ == '__main__':
    test()
